﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System;

namespace HeuristicLab.BioBoost.Utils {

  /// <summary>
  /// Wraps a field into a local variable inside the constructor. This evades
  /// object graph traversal as the actual value is no longer discoverable by
  /// reflection. Use with Caution!
  /// </summary>
  /// <typeparam name="T">The Type of the value to be wrapped.</typeparam>
  [StorableClass]
  public class Closure<T> : IDeepCloneable {

    public enum Encapsulation { Cloned, Referenced };

    [Storable]
    private readonly Encapsulation encapsulation;

    [Storable]
    public T Value { get { return Getter(); } set { Setter(value); } }

    public Func<T> Getter;
    public Action<T> Setter;

    [StorableConstructor]
    protected Closure(bool isDeserializing) {
      T value = default(T);
      Getter = () => value;
      Setter = v => value = v;
    }
    protected Closure(Closure<T> orig, Cloner cloner) {
      T value = default(T);
      Getter = () => value;
      Setter = v => value = v;
      encapsulation = orig.encapsulation;
      cloner.RegisterClonedObject(orig, this);
      if (encapsulation == Encapsulation.Referenced) {
        Value = orig.Value;
      } else {
        var cloneable = orig.Value as IDeepCloneable;
        if (cloneable != null)
          Value = (T) cloner.Clone(cloneable);
        else
          Value = orig.Value;
      }
    }
    public Closure(Encapsulation encapsulation) {
      this.encapsulation = encapsulation;
      T value = default(T);
      Getter = () => value;
      Setter = v => value = v;
    }

    public object Clone() {
      return Clone(new Cloner());
    }

    public IDeepCloneable Clone(Cloner cloner) {
      return new Closure<T>(this, cloner);
    }
  }
}
